<?php

declare(strict_types=1);

namespace Dedi\SyliusPayboxPlugin\Payum;

use Dedi\SyliusPayboxPlugin\Model\Iso3166Country;
use Payum\Core\Bridge\Spl\ArrayObject;
use Payum\Core\Reply\HttpPostRedirect;
use Payum\Core\Security\GenericTokenFactoryInterface;
use Sylius\Bundle\PayumBundle\Model\PaymentSecurityTokenInterface;
use Sylius\Component\Core\Model\PaymentInterface;
use Symfony\Component\String\UnicodeString;

final class PayboxApi implements PayboxParams
{
    public const CODE = 'PAYBOX';

    public const ACCEPTED_STATUS = '00000';

    public const FAILED_STATUS = ['00100', '00199'];

    private array $options = [];

    public function __construct(array $options)
    {
        $this->options = $options;
    }

    public function isAccepted(string $status): bool
    {
        return $status === self::ACCEPTED_STATUS;
    }

    public function isFailed(string $status): bool
    {
        return (int) self::FAILED_STATUS[0] <= (int) $status && (int) self::FAILED_STATUS[1] >= (int) $status;
    }

    public function convertPayment(PaymentInterface $payment, PaymentSecurityTokenInterface $token, GenericTokenFactoryInterface $tokenFactory): array
    {
        $order = $payment->getOrder();
        $details = $payment->getDetails();

        $details = ArrayObject::ensureArrayObject($details);
        $details[self::PBX_TOTAL] = $order->getTotal();
        $details[self::PBX_DEVISE] = self::PBX_DEVISE_EURO;
        $details[self::PBX_CMD] = $order->getNumber();
        $details[self::PBX_PORTEUR] = $order->getCustomer()->getEmail();

        $details[self::PBX_EFFECTUE] = $token->getTargetUrl();
        $details[self::PBX_ANNULE] = $token->getTargetUrl();
        $details[self::PBX_REFUSE] = $token->getTargetUrl();
        $details[self::PBX_TYPECARTE] = 'CB';

        $quantity = $order->countItems();

        $details[self::PBX_SHOPPINGCART] = sprintf(
            '<?xml version="1.0" encoding="utf-8"?><shoppingcart><total><totalQuantity>%d</totalQuantity></total></shoppingcart>',
            (int) $quantity
        );

        $shippingAddress = $order->getShippingAddress();
        $firstName = $this->formatTextValue($shippingAddress->getFirstName(), 'ANP', 30);
        $lastName = $this->formatTextValue($shippingAddress->getLastName(), 'ANP', 30);
        $street = $this->formatTextValue($shippingAddress->getStreet(), 'AN', 50);
        $street2 = $this->formatTextValue(substr($shippingAddress->getStreet(), 50) ?: '', 'AN', 50);
        $zipCode = $this->formatTextValue($shippingAddress->getPostcode(), 'AN', 16);
        $city = $this->formatTextValue($shippingAddress->getCity(), 'AN', 50);
        $countryCode = Iso3166Country::getNumericCode($shippingAddress->getCountryCode(), Iso3166Country::CODES['FR']);

        $details[self::PBX_BILLING] = sprintf(
            '<?xml version="1.0" encoding="utf-8"?><Billing><Address><FirstName>%s</FirstName><LastName>%s</LastName><Address1>%s</Address1><Address2>%s</Address2><ZipCode>%s</ZipCode><City>%s</City><CountryCode>%s</CountryCode></Address></Billing>',
            $firstName,
            $lastName,
            $street,
            $street2,
            $zipCode,
            $city,
            $countryCode
        );

        // Prevent duplicated payment error
        if (strpos($token->getGatewayName(), 'sandbox') !== false) {
            $details[self::PBX_CMD] = sprintf('%s-%d', $details[self::PBX_CMD], time());
        }

        if (!isset($details[PayboxParams::PBX_REPONDRE_A])) {
            $notifyToken = $tokenFactory->createNotifyToken($token->getGatewayName(), $payment);
            $details[PayboxParams::PBX_REPONDRE_A] = $notifyToken->getTargetUrl();
        }

        return (array) $details;
    }

    public function doPayment(array $details): void
    {
        $details[self::PBX_SITE] = $this->options['site'];
        $details[self::PBX_RANG] = $this->options['rang'];
        $details[self::PBX_IDENTIFIANT] = $this->options['identifiant'];
        $details[self::PBX_HASH] = $this->options['hash'];
        $details[self::PBX_SOURCE] = $this->isMobileBrowser() ? self::PBX_SOURCE_MOBILE : self::PBX_SOURCE_DESKTOP;
        $details[self::PBX_RETOUR] = self::PBX_RETOUR_VALUE;
        $details[self::PBX_HMAC] = strtoupper($this->computeHmac($this->options['hmac'], (array) $details));

        $authorizeTokenUrl = $this->getApiEndpoint();

        throw new HttpPostRedirect($authorizeTokenUrl, (array) $details);
    }

    private function getApiEndpoint(): string
    {
        $servers = $this->options['sandbox'] ? self::SERVERS_PREPROD : self::SERVERS_PROD;

        //TODO: add choice for paybox payment page (iframe, mobile or classic)
        $endpoint = self::URL_CLASSIC;

        // Test if paybox server is available
        // otherwise, use fallback url
        foreach ($servers as $server) {
            $doc = new \DOMDocument();
            $doc->loadHTMLFile('https://' . $server . '/load.html');
            $element = $doc->getElementById('server_status');

            if ($element && 'OK' == $element->textContent) {
                return 'https://' . $server . '/' . $endpoint;
            }
        }

        throw new \RuntimeException('Paybox endpoint server not available.');
    }

    private function computeHmac(string $hmac, array $fields)
    {
        // Si la clé est en ASCII, On la transforme en binaire
        $binKey = pack('H*', $hmac);
        $msg = $this->stringify($fields);
        $string = strtoupper(hash_hmac($fields[self::PBX_HASH], $msg, $binKey));

        return $string;
    }

    private function stringify(array $array): string
    {
        $result = [];
        foreach ($array as $key => $value) {
            $result[] = sprintf('%s=%s', $key, $value);
        }

        return implode('&', $result);
    }

    public function getOptions(): array
    {
        return $this->options;
    }

    private function isMobileBrowser(): bool
    {
        $useragent = $_SERVER['HTTP_USER_AGENT'];

        return preg_match('/(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows ce|xda|xiino/i', $useragent) || preg_match('/1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i', substr($useragent, 0, 4));
    }

    /**
     * AN : Alphanumerical without special characters
     * ANP : Alphanumerical with spaces and special characters
     * ANS : Alphanumerical with special characters
     * N : Numerical only
     * A : Alphabetic only
     */
    private function formatTextValue($value, $type, $maxLength = null)
    {
        switch ($type) {
            default:
            case 'AN':
                $value = (string) (new UnicodeString($value))->ascii();

                break;
            case 'ANP':
                $value = (string) (new UnicodeString($value))->ascii();
                $value = preg_replace('/[^-. a-zA-Z0-9]/', '', $value);

                break;
            case 'ANS':
                break;
            case 'N':
                $value = preg_replace('/[^0-9.]/', '', $value);

                break;
            case 'A':
                $value = (string) (new UnicodeString($value))->ascii();
                $value = preg_replace('/[^A-Za-z]/', '', $value);

                break;
        }
        // Remove carriage return characters
        $value = trim(preg_replace("/\r|\n/", '', $value));

        // Cut the string when needed
        if (!empty($maxLength) && is_numeric($maxLength) && $maxLength > 0) {
            if (function_exists('mb_strlen') && mb_strlen($value) > $maxLength) {
                $value = mb_substr($value, 0, $maxLength);
            } elseif (strlen($value) > $maxLength) {
                $value = substr($value, 0, $maxLength);
            }
        }

        return $value;
    }
}
